﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace vamtaco
{
    // IMPORTANT - DO NOT make custom enums. The dynamic C# complier crashes Unity when it encounters these for
    // some reason

    // IMPORTANT - DO NOT OVERRIDE Awake() as it is used internally by MVRScript - instead use Init() function which
    // is called right after creation
    public class VRHandsFingering : MVRScript
    {
        //Finger Node Format
        //R_Finger_Middle_A
        //R_Finger_Middle_B
        //R_Finger_Middle_C

        GameObject playerHands;

        string[] fingers = new string[] 
        {
            "Thumb", "Index", "Middle", "Ring", "Pinky"
        };

        bool _leftStateResetButtonPressed = false;

        bool _leftStateChangeButtonPressed = false;
        int _leftHandStateIndex = -1;

        bool _rightStateResetButtonPressed = false;

        bool _rightStateChangeButtonPressed = false;
        int _rightHandStateIndex = -1;

        Dictionary<int, string> handStates = new Dictionary<int, string>();

        public override void Init()
        {
            try
            {
                handStates.Add(0, "Index");
                handStates.Add(1, "Middle");
                handStates.Add(2, "Ring");
                handStates.Add(3, "Pinky");
                handStates.Add(4, "Thumb");
                handStates.Add(5, "ThumbIndex");
                handStates.Add(6, "IndexMiddle");
                handStates.Add(7, "IndexMiddleRing");
                handStates.Add(8, "IndexMiddlePinky");
                handStates.Add(9, "IndexRingPinky");
                handStates.Add(10, "Fist");
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // Start is called once before Update or FixedUpdate is called and after Init()
        void Start()
        {
            try
            {
                SuperController.singleton.commonHandModelControl.useCollision = true;
                SuperController.singleton.allowGrabPlusTriggerHandToggle = false;

                playerHands = SuperController.singleton.commonHandModelControl.gameObject;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void ChangeHandState(string hand, int state)
        {
            switch(state)
            {
                case 0:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 1:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 2:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 3:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 4:
                    SingleFinger(hand, handStates[state]);
                    break;
                case 5:
                    ThumbIndex(hand);
                    break;
                case 6:
                    IndexMiddle(hand);
                    break;
                case 7:
                    IndexMiddleRing(hand);
                    break;
                case 8:
                    IndexMiddlePinky(hand);
                    break;
                case 9:
                    IndexRingPinky(hand);
                    break;
                case 10:
                    MakeFist(hand);
                    break;
            }
        }

        private void CloseFinger(string hand, string finger)
        {
            playerHands
                .GetComponentsInChildren<Transform>()
                    .Where(t => t.name.Contains($"{hand}_Finger_{finger}")).ToList()
                    .ForEach(t =>
                    {
                        float a = 3;
                        float b = 35;
                        float c = 35;

                        if(hand == "L")
                        {
                            a *= -1;
                            b *= -1;
                            c *= -1;
                        }

                        if(t.name.Contains("_A"))
                            t.Rotate(Vector3.forward, a);
                        else if (t.name.Contains("_B"))
                            t.Rotate(Vector3.forward, b);
                        else
                            t.Rotate(Vector3.forward, c);
                    });
        }

        private void MakeFist(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Index");
            CloseFinger(hand, "Middle");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");
        }

        private void SingleFinger(string hand, string finger)
        {
            foreach(var f in fingers)
            {
                if (f != finger)
                    CloseFinger(hand, f);
            }
        }

        private void ThumbIndex(string hand)
        {
            CloseFinger(hand, "Middle");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");
        }

        private void IndexMiddle(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Ring");
            CloseFinger(hand, "Pinky");

            //Experimental:  Currently has major collision issues and is basically not useable
            //playerHands
            //        .GetComponentsInChildren<Transform>()
            //        .Single(t => t.name.Contains($"{hand}_Finger_Middle_A"))
            //        .Rotate(Vector3.left, 4f);

            //playerHands
            //    .GetComponentsInChildren<Transform>()
            //    .Single(t => t.name.Contains($"{hand}_Finger_Index_A"))
            //    .Rotate(Vector3.right, 5f);
        }

        private void IndexMiddleRing(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Pinky");
        }

        private void IndexMiddlePinky(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Ring");
        }

        private void IndexRingPinky(string hand)
        {
            CloseFinger(hand, "Thumb");
            CloseFinger(hand, "Middle");
        }

        private void HandleUserInputStateChangeButton(ref bool stateChangeButtonPressed, ref int handStateIndex)
        {
            if(!stateChangeButtonPressed)
            {
                if (handStateIndex != handStates.Count - 1)
                {
                    handStateIndex++;
                }
                else
                {
                    handStateIndex = -1;
                }
            }
            
            stateChangeButtonPressed = true;
        }

        private void HandleUserInputStateResetButton(ref bool stateChangeButtonPressed, ref int handStateIndex)
        {
            if (!stateChangeButtonPressed)
            {
                handStateIndex = -1;
            }

            stateChangeButtonPressed = true;
        }

        // Update is called with each rendered frame by Unity
        void Update()
        {
            try
            {
                if(SuperController.singleton.GetLeftSelect())
                    HandleUserInputStateChangeButton(ref _leftStateChangeButtonPressed, ref _leftHandStateIndex);

                if (SuperController.singleton.GetGrabNavigateLeft())
                    HandleUserInputStateResetButton(ref _leftStateResetButtonPressed, ref _leftHandStateIndex);

                if (SuperController.singleton.GetRightSelect())
                    HandleUserInputStateChangeButton(ref _rightStateChangeButtonPressed, ref _rightHandStateIndex);

                if (SuperController.singleton.GetGrabNavigateRight())
                    HandleUserInputStateResetButton(ref _rightStateResetButtonPressed, ref _rightHandStateIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // FixedUpdate is called with each physics simulation frame by Unity
        void FixedUpdate()
        {
            try
            {
                if (_leftStateChangeButtonPressed)
                    _leftStateChangeButtonPressed = false;

                if(_leftStateResetButtonPressed)
                    _leftStateResetButtonPressed = false;
                
                if (_rightStateChangeButtonPressed)
                    _rightStateChangeButtonPressed = false;

                if (_rightStateResetButtonPressed)
                    _rightStateResetButtonPressed = false;

                ChangeHandState("L", _leftHandStateIndex);
                ChangeHandState("R", _rightHandStateIndex);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        // OnDestroy is where you should put any cleanup
        // if you registered objects to supercontroller or atom, you should unregister them here
        void OnDestroy()
        {
        }
    }
}
